/**
 * Copyright (c) 2024 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include "DistrhoUI.hpp"
#include "ResizeHandle.hpp"

START_NAMESPACE_DISTRHO

// --------------------------------------------------------------------------------------------------------------------
enum HeavyParams {HIGH,HIGH_FEEDBACK,HIGH_INTENSITY,HIGH_MIX,HIGH_SPEED,LOW,LOW_FEEDBACK,LOW_INTENSITY,LOW_MIX,LOW_SPEED,MID,MID_FEEDBACK,MID_FREQ,MID_INTENSITY,MID_MIX,MID_SPEED,
};

class ImGuiPluginUI : public UI
{
    
    float fhigh = 0.0f;
    float fhigh_feedback = 0.0f;
    float fhigh_intensity = 20.0f;
    float fhigh_mix = 50.0f;
    float fhigh_speed = 2.0f;
    float flow = 0.0f;
    float flow_feedback = 0.0f;
    float flow_intensity = 20.0f;
    float flow_mix = 50.0f;
    float flow_speed = 2.0f;
    float fmid = 0.0f;
    float fmid_feedback = 0.0f;
    float fmid_freq = 1337.0f;
    float fmid_intensity = 20.0f;
    float fmid_mix = 50.0f;
    float fmid_speed = 2.0f;

    ResizeHandle fResizeHandle;

    // ----------------------------------------------------------------------------------------------------------------

public:
   /**
      UI class constructor.
      The UI should be initialized to a default state that matches the plugin side.
    */
    ImGuiPluginUI()
        : UI(DISTRHO_UI_DEFAULT_WIDTH, DISTRHO_UI_DEFAULT_HEIGHT),
          fResizeHandle(this)
    {
        setGeometryConstraints(DISTRHO_UI_DEFAULT_WIDTH, DISTRHO_UI_DEFAULT_HEIGHT, true);

        // hide handle if UI is resizable
        if (isResizable())
            fResizeHandle.hide();
    }

protected:
    // ----------------------------------------------------------------------------------------------------------------
    // DSP/Plugin Callbacks

   /**
      A parameter has changed on the plugin side.@n
      This is called by the host to inform the UI about parameter changes.
    */
    void parameterChanged(uint32_t index, float value) override
    {
        switch (index) {
            case HIGH:
                fhigh = value;
                break;
            case HIGH_FEEDBACK:
                fhigh_feedback = value;
                break;
            case HIGH_INTENSITY:
                fhigh_intensity = value;
                break;
            case HIGH_MIX:
                fhigh_mix = value;
                break;
            case HIGH_SPEED:
                fhigh_speed = value;
                break;
            case LOW:
                flow = value;
                break;
            case LOW_FEEDBACK:
                flow_feedback = value;
                break;
            case LOW_INTENSITY:
                flow_intensity = value;
                break;
            case LOW_MIX:
                flow_mix = value;
                break;
            case LOW_SPEED:
                flow_speed = value;
                break;
            case MID:
                fmid = value;
                break;
            case MID_FEEDBACK:
                fmid_feedback = value;
                break;
            case MID_FREQ:
                fmid_freq = value;
                break;
            case MID_INTENSITY:
                fmid_intensity = value;
                break;
            case MID_MIX:
                fmid_mix = value;
                break;
            case MID_SPEED:
                fmid_speed = value;
                break;
            
            default: return;
        }
    
        repaint();
    }

    // ----------------------------------------------------------------------------------------------------------------
    // Widget Callbacks

   /**
      ImGui specific onDisplay function.
    */
    void onImGuiDisplay() override
    {
        const float width = getWidth();
        const float height = getHeight();
        const float margin = 20.0f * getScaleFactor();

        ImGui::SetNextWindowPos(ImVec2(margin, margin));
        ImGui::SetNextWindowSize(ImVec2(width - 2 * margin, height - 2 * margin));

        if (ImGui::Begin("WSTD FL3NGR", nullptr, ImGuiWindowFlags_NoResize + ImGuiWindowFlags_NoCollapse))
        {
            if (ImGui::SliderFloat("High", &fhigh, -15.0f, 15.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH, true);
                }
                setParameterValue(HIGH, fhigh);
            }
    
            if (ImGui::SliderFloat("High Feedback", &fhigh_feedback, -100.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_FEEDBACK, true);
                }
                setParameterValue(HIGH_FEEDBACK, fhigh_feedback);
            }
    
            if (ImGui::SliderFloat("High Intensity", &fhigh_intensity, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_INTENSITY, true);
                }
                setParameterValue(HIGH_INTENSITY, fhigh_intensity);
            }
    
            if (ImGui::SliderFloat("High Mix", &fhigh_mix, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_MIX, true);
                }
                setParameterValue(HIGH_MIX, fhigh_mix);
            }
    
            if (ImGui::SliderFloat("High Speed", &fhigh_speed, 0.0f, 20.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(HIGH_SPEED, true);
                }
                setParameterValue(HIGH_SPEED, fhigh_speed);
            }
    
            if (ImGui::SliderFloat("Low", &flow, -15.0f, 15.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW, true);
                }
                setParameterValue(LOW, flow);
            }
    
            if (ImGui::SliderFloat("Low Feedback", &flow_feedback, -100.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_FEEDBACK, true);
                }
                setParameterValue(LOW_FEEDBACK, flow_feedback);
            }
    
            if (ImGui::SliderFloat("Low Intensity", &flow_intensity, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_INTENSITY, true);
                }
                setParameterValue(LOW_INTENSITY, flow_intensity);
            }
    
            if (ImGui::SliderFloat("Low Mix", &flow_mix, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_MIX, true);
                }
                setParameterValue(LOW_MIX, flow_mix);
            }
    
            if (ImGui::SliderFloat("Low Speed", &flow_speed, 0.0f, 20.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(LOW_SPEED, true);
                }
                setParameterValue(LOW_SPEED, flow_speed);
            }
    
            if (ImGui::SliderFloat("Mid", &fmid, -15.0f, 15.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID, true);
                }
                setParameterValue(MID, fmid);
            }
    
            if (ImGui::SliderFloat("Mid Feedback", &fmid_feedback, -100.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_FEEDBACK, true);
                }
                setParameterValue(MID_FEEDBACK, fmid_feedback);
            }
    
            if (ImGui::SliderFloat("Mid Freq", &fmid_freq, 313.3f, 5705.6f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_FREQ, true);
                }
                setParameterValue(MID_FREQ, fmid_freq);
            }
    
            if (ImGui::SliderFloat("Mid Intensity", &fmid_intensity, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_INTENSITY, true);
                }
                setParameterValue(MID_INTENSITY, fmid_intensity);
            }
    
            if (ImGui::SliderFloat("Mid Mix", &fmid_mix, 0.0f, 100.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_MIX, true);
                }
                setParameterValue(MID_MIX, fmid_mix);
            }
    
            if (ImGui::SliderFloat("Mid Speed", &fmid_speed, 0.0f, 20.0f))
            {
                if (ImGui::IsItemActivated())
                {
                    editParameter(MID_SPEED, true);
                }
                setParameterValue(MID_SPEED, fmid_speed);
            }
    
            if (ImGui::IsItemDeactivated())
            {
            editParameter(HIGH, false);
            editParameter(HIGH_FEEDBACK, false);
            editParameter(HIGH_INTENSITY, false);
            editParameter(HIGH_MIX, false);
            editParameter(HIGH_SPEED, false);
            editParameter(LOW, false);
            editParameter(LOW_FEEDBACK, false);
            editParameter(LOW_INTENSITY, false);
            editParameter(LOW_MIX, false);
            editParameter(LOW_SPEED, false);
            editParameter(MID, false);
            editParameter(MID_FEEDBACK, false);
            editParameter(MID_FREQ, false);
            editParameter(MID_INTENSITY, false);
            editParameter(MID_MIX, false);
            editParameter(MID_SPEED, false);
            }
        }
        ImGui::End();
    }

    DISTRHO_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(ImGuiPluginUI)
};

// --------------------------------------------------------------------------------------------------------------------

UI* createUI()
{
    return new ImGuiPluginUI();
}

// --------------------------------------------------------------------------------------------------------------------

END_NAMESPACE_DISTRHO